local AddonName, a = ...
if a.BuildFail(50000) then return end
local L = a.Localize
local s = SpellFlashAddon
local x = s.UpdatedVariables

-- what should we return if the target is a training dummy?
a.DummyTTD = 600

-- are we dubugging?
a.Debugging = false

-- this should not change, ever, so we set it here:
local myself

-- One array to hold our AoE targets
local AoETargets = {}

-- This is for tracking damage taken
local DamageTaken = {}
local arrayIndex = 0
local tsIndex = 0
local myLastDamageTaken = GetTime()

-- Two arrays to hold the information we need to calculate the DPS on each mob
local TimeFirstDamaged = {}
local AmountHPLost = {}

-- are we currently watching the combat log?
local IsLogging = false

-- Last time we saw HP changes during the current fight. Used to calculate DPS.
local LastTimestamp = false
local LastAmount = false

-- temporary variables
local GUID
local ts, event, sourceGUID, sourceFlags, destGUID, destFlags
local target_is_outsider
local source_is_outsider

local UnitFlag

local CombatStart


-- combat log events to watch for damage
local DamageLogEvents = {
	"SPELL_DAMAGE",
	"DAMAGE_SHIELD",
	"SPELL_PERIODIC_DAMAGE",
	"SPELL_BUILDING_DAMAGE",
	"RANGE_DAMAGE",
}

-- combat log events to watch for healing
local HealingLogEvents = {
	"SPELL_HEAL",
	"SPELL_PERIODIC_HEAL",
}

-- Something to ease debugging
a.Debug = function(output)
	if a.Debugging then
		a.print(output)
	end
end

-- record incoming damage with the damaged mob's guid in a table
local logDamage = function(...)
	LastTimestamp, _, _, _, _, _, _, GUID, _, UnitFlag, _, _, _, _, LastAmount = ...
	if LastAmount and bit.band(UnitFlag, COMBATLOG_OBJECT_REACTION_MASK) >= 32 then
		if TimeFirstDamaged[GUID] then
			a.Debug("logDamage: We already know this mob and add ".. LastAmount .. " to AmountHPLost.")
			AmountHPLost[GUID] = AmountHPLost[GUID] + LastAmount
		else
			a.Debug("logDamage: We have not seen this mob before and add ".. LastTimestamp .. " to TimeFirstDamaged.")
			TimeFirstDamaged[GUID] = LastTimestamp
			AmountHPLost[GUID] = LastAmount
		end
	else
		if (not LastAmount) then
			a.debug("logDamage: There was no amount of damage associated with this event!", ...)
		end
	end
end

-- record incoming damage with the damaged mob's guid in a table
local logSwing = function(...)
	LastTimestamp, _, _, _, _, _, _, GUID, _, UnitFlag, _, LastAmount = ...
	if LastAmount and bit.band(UnitFlag, COMBATLOG_OBJECT_REACTION_MASK) >= 32 then
		if TimeFirstDamaged[GUID] then
			a.Debug("logSwing: We already know this mob and add ".. LastAmount .. " to AmountHPLost.")
			AmountHPLost[GUID] = AmountHPLost[GUID] + LastAmount
		else
			a.Debug("logSwing: We have not seen this mob before and add ".. LastTimestamp .. " to TimeFirstDamaged.")
			TimeFirstDamaged[GUID] = LastTimestamp
			AmountHPLost[GUID] = LastAmount
		end
	else
		if (not LastAmount) then
			a.debug("logSwing: There was no amount of damage associated with this event!", ...)
		end
	end
end

-- subtract incoming healing from the damaged mob's guid
local logHealing = function(...)
	LastTimestamp, _, _, _, _, _, _, GUID, _, UnitFlag, _, _, _, _, LastAmount = ...
	if LastAmount and bit.band(UnitFlag, COMBATLOG_OBJECT_REACTION_MASK) >= 32 then
		if TimeFirstDamaged[GUID] then
			a.Debug("logHealing: We already know this mob and subtract ".. LastAmount .. " from AmountHPLost.")
			AmountHPLost[GUID] = AmountHPLost[GUID] - LastAmount
		else
			a.Debug("logHealing: We did not see the mob get damaged before.")
		end
	else
		if (not LastAmount) then
			a.print("logHealing: There was no amount of healing associated with this event!", ...)
		end
	end
end

local logDoTs = function(...)
	local _, event, _, _, _, _, _, _, _, _, _, _, spellName = ...
	local name = s.NormalizeName(spellName)
	if (event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_REFRESH") and x.Dot[name] then
		x.Dot[name].ticks = 0
		x.Dot[name].ticking = true
	elseif event == "SPELL_PERIODIC_DAMAGE" and x.Dot[name] then
		x.Dot[name].ticks = (x.Dot[name].ticks or 0) + 1
		x.Dot[name].ticking = true
	elseif event == "SPELL_AURA_REMOVED" and x.Dot[name] then
		x.Dot[name].ticks = 0
		x.Dot[name].ticking = false
	end
end

local combatLogEventHandler = function (event, ...)
	_, event, _, sourceGUID, _, sourceFlags, _, destGUID, _, destFlags = ...
	myself = UnitGUID("player")
	pet = UnitGUID("pet")
	ts = GetTime()
	tsIndex = mod(floor(ts * 10), 100) + 1
	
	target_is_outsider = bit.band(destFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER) > 0
	source_is_outsider = bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER) > 0
	
	if tsIndex > arrayIndex then
	-- the new index comes after the old index. we have to clear all array entries between them.
		for i = arrayIndex + 1, tsIndex, 1 do
			DamageTaken[i] = 0
		end
		arrayIndex = tsIndex
	elseif tsIndex < arrayIndex then
	-- the new index comes before the old index. we have to clear till the end of the array and then
	-- from the start up to the new index.
		for i = arrayIndex + 1, 100, 1 do
			DamageTaken[i] = 0
		end
		for i = 1, tsIndex, 1 do
			DamageTaken[i] = 0
		end
		arrayIndex = tsIndex
	end

	if destGUID == myself and string.find(event, "_DAMAGE") then
		myLastDamageTaken = ts
		if destGUID == myself and (event == "SWING_DAMAGE") then
			amount = select(12, ...)
			DamageTaken[arrayIndex] = DamageTaken[arrayIndex] + amount
		else
			amount = select(15, ...)
			DamageTaken[arrayIndex] = DamageTaken[arrayIndex] + amount
		end
	end
	
	if (sourceGUID == myself or sourceGUID == pet) and (destGUID ~= myself and destGUID ~= pet) and target_is_outsider and (string.find(event, "_DAMAGE") or string.find(event, "_MISSED")) then
		AoETargets[destGUID] = GetTime()
	end
	if event == "UNIT_DIED" or event == "UNIT_DESTROYED" or event == "UNIT_DISSIPATES" then
		AoETargets[destGUID] = nil
	end
	
	if tContains(DamageLogEvents, event) then
		a.Debug(event .. " fired, calling logDamage")
		logDamage(...)
	elseif tContains(HealingLogEvents, event)  then
		a.Debug(event .. " fired, calling logHealing")
		logHealing(...)
	elseif event == "SWING_DAMAGE" then
		logSwing(...)
	end

	if sourceGUID == myself and string.find(event, "SPELL_") then
		logDoTs(...)
	end

end

local clearDamageTaken = function ()
	for i = 1, 100, 1 do
		DamageTaken[i] = 0
	end
end
clearDamageTaken()

-- start the combat log (when the player enters combat)
local enterCombat = function(...)
	s.RegisterModuleEvent(AddonName, "COMBAT_LOG_EVENT_UNFILTERED", combatLogEventHandler)
	IsLogging = true
	CombatStart = GetTime()
	a.Debug("enterCombat: Started watching for COMBAT_LOG_EVENT_UNFILTERED.")
end

-- stop the combat log (when the player leaves combat)
local leaveCombat = function(...)
	s.UnregisterModuleEvent(AddonName, "COMBAT_LOG_EVENT_UNFILTERED")
	LastTimestamp = false
	LastAmount = false
	wipe(TimeFirstDamaged)
	wipe(AmountHPLost)
	IsLogging = false
	CombatStart = 0
	clearDamageTaken()
	a.Debug("leaveCombat: Stopped watching for COMBAT_LOG_EVENT_UNFILTERED.")
end

-- the public function that this is all about
s.TimeToDie = function(unit)
	if not unit then
		unit = "target"
	end
	a.Debug('s.TimeToDie("'..unit..'") called.')
	if  IsLogging then
		GUID = UnitGUID(unit)
		if GUID then
			if s.Dummy(unit) then
				a.Debug('s.TimeToDie: Training Dummy detected. Returning '..a.DummyTTD)
				return a.DummyTTD
			end
			if UnitHealth(unit) == 0 then
				return 0
			elseif TimeFirstDamaged[GUID] and LastTimestamp then
				
				return UnitHealth(unit)  * (LastTimestamp - TimeFirstDamaged[GUID]) / AmountHPLost[GUID]
			end
		else
			a.Debug('s.TimeToDie: UnitGUID returned nil. Exiting.')
		end
	else
		a.Debug('s.TimeToDie: not logging. Exiting.')
	end
end

-- register events
s.RegisterModuleEvent(AddonName, "PLAYER_REGEN_ENABLED", leaveCombat)
s.RegisterModuleEvent(AddonName, "PLAYER_REGEN_DISABLED", enterCombat)
s.RegisterModuleEvent(AddonName, "PLAYER_LOGIN", leaveCombat)


s.Time = function()
	if CombatStart == 0 then
		return 0
	end
	return GetTime() - CombatStart
end

s.SpellTargets = function()
	local now = GetTime()
	local i = 0
	foreach(AoETargets, function(k, v)
		if v and now - v > 3 then
			AoETargets[k] = nil
		else
			i = i + 1
		end
	end)
	return i
end

s.IncomingDamage = function(ms)
	local segments = floor(ms / 100)
	local lowStart = arrayIndex - segments + 1
	local highStart = 0
	if lowStart < 1 then
		highStart = 100 + lowStart
		lowStart = 1
	end
	local i
	local ret = 0
	for i = lowStart, arrayIndex, 1 do
		ret = ret + DamageTaken[i]
	end
	if highStart ~= 0 then
		for i = highStart, 100, 1 do
			ret = ret + DamageTaken[i]
		end
	end
	return ret
end
