local AddonName, a = ...
if a.BuildFail(50000) then return end
local s = SpellFlashAddon
local x = s.UpdatedVariables

x.PrevGCD = {}
x.PrevGCD[1] = {}
x.PrevGCD[2] = {}
x.PrevOffGCD = {}

local lastGCD = 0
local lastStartSpell = ""

local clearOffGCD = function()
	for key, val in pairs(x.PrevOffGCD) do
		x.PrevOffGCD[key] = nil
	end
end

-- This handles spells with a cast time that start a GCD
a.UNIT_SPELLCAST_START = function(event, unit, name, _, _, spellId)
	if unit == "player" then
		name = s.NormalizeName(name)
		-- This gets us the GCD
		local startGCD, durationGCD = GetSpellCooldown(61304)
		-- When this event fires, the spell's CD (if it has one) has not registered yet. If the spell has a CD at all,
		-- then it's the GCD at this time
		local startSpell, durationSpell = GetSpellCooldown(spellId)
		-- If the spell starts a GCD, then its start time and the GCD's start time are identical
		if (startSpell ~= lastGCD) and startGCD == startSpell and durationGCD > 0 and durationSpell > 0 then
			-- We have to remember that this is a spell with cast time. Otherwise it can get confusing if a cast-time-spell
			-- and an instant-cast-spell successfully finish at the same time
			lastStartSpell = name
			-- I am unsure whether this is necessary. Starting a spell may proc other "spells" whose cooldowns all have the same start time.
			-- We do not want to register those and they usually fire after the spell triggering them...
			lastGCD = startSpell
			x.PrevGCD[2] = x.PrevGCD[1]
			x.PrevGCD[1] = {}
			x.PrevGCD[1][name] = true
			-- The way Simulationcraft seems to work, we are not interested in off-GCD spells after we start a new GCD:
			clearOffGCD()
		end
	end

	-- Special handler for Water Jet, should probably be moved to the mage module
	if unit == "pet" then
		name = s.NormalizeName(name)
		if name ~= "water_jet" then
			return
		end
		clearOffGCD()
		x.PrevOffGCD[name] = true
	end
end
s.RegisterModuleEvent(AddonName, "UNIT_SPELLCAST_START", a.UNIT_SPELLCAST_START)

-- This is for instant spells that start a GCD and also for spells off the GCD
a.UNIT_SPELLCAST_SUCCEEDED = function(event, unit, name, _, _, spellId)
	if unit == "player" then
		name = s.NormalizeName(name)
		if name == lastStartSpell then
			return
		end
		-- This gets us the GCD
		local startGCD, durationGCD = GetSpellCooldown(61304)
		-- When this event fires, the spell's CD (if it has one) has not registered yet. If the spell has a CD at all,
		-- then it's the GCD at this time
		local startSpell, durationSpell = GetSpellCooldown(spellId)
		-- If the spell starts a GCD, then its start time and the GCD's start time are identical
		if startSpell ~= lastGCD and startGCD == startSpell and durationGCD > 0 and durationSpell > 0 then
			-- I am unsure whether this is necessary. Starting a spell may proc other "spells" whose cooldowns all have the same start time.
			-- We do not want to register those and they usually fire after the spell triggering them...
			lastGCD = startGCD
			x.PrevGCD[2] = x.PrevGCD[1]
			x.PrevGCD[1] = {}
			x.PrevGCD[1][name] = true
			-- The way Simulationcraft seems to work, we are not interested in off-GCD spells after we start a new GCD:
			clearOffGCD()
		elseif durationSpell == 0 then
			-- if we are here, then the "spell" does not have a GCD. It does not matter whether it has a cooldown itself, as this
			-- cooldown would not have registered yet, when this event fires.
        	local baseCooldown = GetSpellBaseCooldown(spellId) or 0
        	-- We use the base cooldown to find out whether the spell is a spell that was actually cast by the player:
	        if baseCooldown > 0 then
				clearOffGCD()
				x.PrevOffGCD[name] = true
			end
		end
	end
end
s.RegisterModuleEvent(AddonName, "UNIT_SPELLCAST_SUCCEEDED", a.UNIT_SPELLCAST_SUCCEEDED)

-- We use this to clear the previous GCD if a spell with cast time has been stopped or interrupted:
a.RESET_PREV_GCD = function(event, unit, name, _, _, spellId)
	if unit == "player" then
		name = s.NormalizeName(name)
		if lastStartSpell == name then
			x.PrevGCD[1] = {}
		end
		lastStartSpell = ""
	end
end
s.RegisterModuleEvent(AddonName, "UNIT_SPELLCAST_STOP", a.RESET_PREV_GCD)
s.RegisterModuleEvent(AddonName, "UNIT_SPELLCAST_INTERRUPTED", a.RESET_PREV_GCD)
